from upk2esphome.config import ConfigData
from upk2esphome.generator import invert, pull
from upk2esphome.opts import Opts
from upk2esphome.result import YamlResult


def generate(yr: YamlResult, config: ConfigData, opts: Opts):
    config = config.upk or {}
    in_keys = ["ir", "infrr"]
    out_keys = ["infre"]

    for key in in_keys:
        ir_pin = config.get(f"{key}")

        # keys currently not understood:
        # irfunc
        # irnightt
        # irstep
        # irk*val, irk*fun - presumably these map ir values received to specific functions (light on, off, colours, etc.)

        if ir_pin is None:
            continue

        yr.log(f"Remote receiver: pin P{ir_pin}")
        receiver = {
            "pin": f"P{ir_pin}",
        }
        pull(receiver, True)
        receiver["_1"] = "dump: all"
        yr.data["remote_receiver"] = receiver

    for key in out_keys:
        ir_pin = config.get(f"{key}")

        if ir_pin is None:
            continue

        yr.log(f"Remote transmitter: pin P{ir_pin}")
        transmitter = {
            "pin": f"P{ir_pin}",
        }
        transmitter["carrier_duty_percent"] = "50%"
        yr.data["remote_transmitter"] = transmitter



#  Copyright (c) Kuba Szczodrzyński 2023-9-4.

from upk2esphome.config import ConfigData
from upk2esphome.datapoints import Datapoint, convert_all
from upk2esphome.opts import Opts
from upk2esphome.result import YamlResult


def generate(yr: YamlResult, config: ConfigData, opts: Opts):
    if not config.is_tuya_mcu:
        return

    yr.log("Found TuyaMCU device")

    baud_rate = config.uart_config.get("uart_baud", None) or config.uart_config.get(
        "baud", None
    )
    if not baud_rate:
        baud_rate = "REPLACEME"
        yr.warn(
            "TuyaMCU baud rate not found!\n"
            "You need to adjust it manually: in most cases, use either 9600 or 115200"
        )

    yr.data["uart"] = {
        "rx_pin": "RX1",
        "tx_pin": "TX1",
        "baud_rate": baud_rate,
    }
    yr.data["tuya"] = {}
    yr.found = True

    if not config.model:
        yr.warn("No schema model for TuyaMCU - can't process datapoints")
        yr.needs_tuyamcu_model = True
        return

    model = config.model
    model_id = model.get("modelId", None)
    schema_id = config.schema_id
    services = model.get("services", None)
    if not model_id or not isinstance(services, list):
        yr.warn("Invalid schema model structure: missing 'modelId' or 'services'")
        yr.needs_tuyamcu_model = True
        return

    if schema_id and model_id != schema_id:
        yr.warn(
            f"Downloaded schema ID ({model_id}) "
            f"doesn't match device schema ID ({schema_id})"
        )

    yr.data["tuya"] = {
        "_1": f"DPIDs processed from schema model: {model_id}",
    }

    # mark 'found' only if supported DP component was processed
    yr.found = False
    dps: list[Datapoint] = []
    for service in services:
        for dp_data in service.get("properties", []):
            dps.append(Datapoint.build(dp_data))

    convert_all(yr, config, opts, dps)



#  Copyright (c) Kuba Szczodrzyński 2023-4-21.

import json
import re
from typing import Union

from upk2esphome.config import ConfigData
from upk2esphome.opts import Opts
from upk2esphome.result import YamlResult


def invert(data: dict, inverted: bool, key: str = "pin"):
    if inverted:
        data[key] = {
            "number": data[key],
            "inverted": True,
        }


def pull(data: dict, inverted: bool, key: str = "pin"):
    if inverted:
        data[key] = {
            "number": data[key],
            "inverted": True,
            "mode": "INPUT_PULLUP",
        }
    else:
        data[key] = {
            "number": data[key],
            "mode": "INPUT_PULLDOWN",
        }


def parse_user_param_key(value: str) -> dict | None:
    value = re.sub(r"([^{}\[\]:,]+)", r'"\1"', value)
    value = re.sub(r'"([1-9][0-9]*|0)"', r"\1", value)
    value = re.sub(",}", "}", value)
    value = json.loads(value)
    return value


def parse_raw_data(raw_value: Union[str, dict]) -> dict:
    loaders = [
        json.loads,
        parse_user_param_key,
        dict,
    ]
    if isinstance(raw_value, str):
        raw_value = raw_value.strip()
    value: dict | None = None
    for loader in loaders:
        try:
            value = loader(raw_value)
            break
        except Exception:
            pass
    if not isinstance(value, dict):
        raise ValueError("Input data parsing error")
    return value


def parse_input(
    raw_data: Union[str, dict],
    raw_extras: Union[str, dict],
) -> ConfigData:
    data = parse_raw_data(raw_data)
    extras = parse_raw_data(raw_extras)
    return ConfigData.build(data, extras)


def upk2esphome(
    raw_data: Union[str, dict],
    opts: Opts,
    raw_extras: Union[str, dict] = None,
) -> YamlResult:
    from .parts import bulb, module, monitor, netled, remote, static, switch, tuyamcu

    parts = [module, static, bulb, switch, netled, remote, monitor, tuyamcu]
    yr = YamlResult()

    try:
        # detect input data type
        config = parse_input(raw_data, raw_extras or {})
    except ValueError as e:
        yr.error(str(e))
        return yr

    for part in parts:
        part.generate(yr, config, opts)

    if not config.upk and not config.is_tuya_mcu:
        # unknown type
        yr.error(
            "The chosen device doesn't contain pin configuration.\n\n"
            "Possible causes:\n"
            "- it has vendor-specific firmware\n"
            "- it uses TuyaMCU (report error if that's the case!)\n\n"
            "Auto-generating ESPHome YAML is not possible."
        )
        yr.data = {}

    if not yr.found and not yr.errors:
        yr.error(
            "No actual components were added!\n"
            "This means that **valid configuration has been found**,\n"
            "but this type of device is not yet supported by this program.\n"
            "This includes for example thermometers, water leak sensors, "
            "or fan controllers."
        )

    if yr.errors:
        yr.logs = []
        yr.warnings = []

    yr.config = config
    return yr



#  Copyright (c) Kuba Szczodrzyński 2023-4-21.

import re

import yaml

from upk2esphome.config import ConfigData


class MyDumper(yaml.Dumper):
    def increase_indent(self, flow=False, **_):
        return super(MyDumper, self).increase_indent(flow, False)


class YamlResult:
    data: dict
    logs: list[str]
    warnings: list[str]
    errors: list[str]
    found: bool
    config: ConfigData | None
    needs_tuyamcu_model: bool

    def __init__(self):
        self.data = {}
        self.logs = []
        self.warnings = []
        self.errors = []
        self.found = False
        self.config = None
        self.needs_tuyamcu_model = False

    @property
    def text(self) -> str:
        text = yaml.dump(self.data, sort_keys=False, Dumper=MyDumper)
        text = re.sub(r"'!(\w+) (.+)'", r"!\1 \2", text)
        text = re.sub(r"\n([a-z])", r"\n\n\1", text)
        text = text.replace("'", '"')
        # generate comments for quoted strings with _\d+ keys
        text = re.sub(r'_\d+: "(.+?)"', r"# \1", text)
        # generate comments for unquoted strings _\d+ keys
        text = re.sub(r"_\d+: (.+?)", r"# \1", text)
        text = text.replace(" {}", "")
        text = text.replace("{}", "")
        return text

    def log(self, text: str):
        self.logs.append(text)

    def warn(self, text: str):
        self.warnings.append(text)

    def error(self, text: str):
        self.errors.append(text)

    def component(self, data: dict):
        if "external_components" not in self.data:
            self.data["external_components"] = []
        self.data["external_components"].append(data)

    def output(self, data: dict):
        if "output" not in self.data:
            self.data["output"] = []
        self.data["output"].append(data)

    def light(self, data: dict):
        if "light" not in self.data:
            self.data["light"] = []
        self.data["light"].append(data)

    def switch(self, data: dict):
        if "switch" not in self.data:
            self.data["switch"] = []
        self.data["switch"].append(data)

    def binary(self, data: dict):
        if "binary_sensor" not in self.data:
            self.data["binary_sensor"] = []
        self.data["binary_sensor"].append(data)

    def sensor(self, data: dict):
        if "sensor" not in self.data:
            self.data["sensor"] = []
        self.data["sensor"].append(data)

    def text_sensor(self, data: dict):
        if "text_sensor" not in self.data:
            self.data["text_sensor"] = []
        self.data["text_sensor"].append(data)

    def button(self, data: dict):
        if "button" not in self.data:
            self.data["button"] = []
        self.data["button"].append(data)

    def number(self, data: dict):
        if "number" not in self.data:
            self.data["number"] = []
        self.data["number"].append(data)

    def select(self, data: dict):
        if "select" not in self.data:
            self.data["select"] = []
        self.data["select"].append(data)

    def tuya_on_datapoint_update(self, data: dict):
        if "tuya" not in self.data:
            self.data["tuya"] = tuya = {}
        else:
            tuya = self.data["tuya"]
        if "on_datapoint_update" not in tuya:
            tuya["on_datapoint_update"] = []
        tuya["on_datapoint_update"].append(data)



#  Copyright (c) Kuba Szczodrzyński 2023-9-5.

from logging import debug
from typing import Callable

import requests
from ltchiptool.gui.work.base import BaseThread

from upk2esphome.const import SCHEMA_PULL_URL, SCHEMA_URL_PREFIX

SCHEMA_PULL_USER = "upk-ltchiptool-v1"
SCHEMA_PULL_PASS = "70b5d7b24cfbba7c0ca831d337c99b37"


class UpkSchemaThread(BaseThread):
    def __init__(
        self,
        device: dict,
        software: dict | None,
        license_: dict | None,
        schema_id: str | None,
        on_success: Callable[[dict], None] = None,
    ):
        super().__init__()
        self.device = device or {}
        self.software = software or {}
        self.license = license_ or {}
        self.schema_id = schema_id
        self.on_success = on_success

    def run_impl(self):
        auth = (SCHEMA_PULL_USER, SCHEMA_PULL_PASS)
        if self.schema_id:
            debug(f"Schema: schema_id={self.schema_id}")
            with requests.get(url=SCHEMA_URL_PREFIX + self.schema_id, auth=auth) as r:
                if r.status_code == 200:
                    debug(f"Schema: found model by ID")
                    self.on_success(r.json())
                    return

        debug(f"Schema: device={self.device}")
        debug(f"Schema: software={self.software}")
        debug(f"Schema: license={self.license}")
        if (
            not self.device.get("firmwareKey", None)
            and not self.device.get("productKey", None)
            and not self.device.get("factoryPin", None)
        ):
            raise ValueError("Missing FK/PK/FC")
        if self.license.get("uuid", None) and not self.license.get("authKey", None):
            raise ValueError("Missing License Auth Key")
        if not self.license.get("uuid", None) and self.license.get("authKey", None):
            raise ValueError("Missing License UUID")

        data = dict(
            device=self.device,
            software=self.software,
            license=self.license,
        )
        with requests.post(url=SCHEMA_PULL_URL, json=data, auth=auth) as r:
            if r.status_code != 200:
                raise RuntimeError(
                    r.json().get("message", f"Response status code {r.status_code}"),
                )
            self.on_success(r.json())



#  Copyright (c) Kuba Szczodrzyński 2023-9-4.

from dataclasses import dataclass
from enum import Enum
from typing import Literal


@dataclass
class ConfigData:
    type: "Type"
    data: dict
    extras: dict

    class Type(Enum):
        CLOUDCUTTER = "Cloudcutter JSON"
        STORAGE = "Storage JSON"
        RAW = "Raw UPK"

    @staticmethod
    def build(data: dict, extras: dict) -> "ConfigData":
        if "schemas" in data and "profiles" in data:
            type = ConfigData.Type.CLOUDCUTTER
        elif "gw_bi" in data:
            type = ConfigData.Type.STORAGE
        elif "Jsonver" in data or "jv" in data or "crc" in data:
            type = ConfigData.Type.RAW
            data = dict(sorted(data.items()))
        else:
            raise ValueError("Couldn't recognize input data")
        return ConfigData(type, data, extras)

    @property
    def is_tuya_mcu(self) -> bool:
        match self.type:
            case ConfigData.Type.CLOUDCUTTER:
                slug = self.profile_slug or ""
                return slug.startswith("bk7231") and "common" in slug
            case ConfigData.Type.STORAGE:
                return "baud_cfg" in self.data or "uart_adapt_params" in self.data
            case ConfigData.Type.RAW:
                return False

    @property
    def upk(self) -> dict:
        match self.type:
            case ConfigData.Type.CLOUDCUTTER:
                return self.data.get("device_configuration", {})
            case ConfigData.Type.STORAGE:
                return self.data.get("user_param_key", {})
            case ConfigData.Type.RAW:
                return self.data

    @property
    def uart_config(self) -> dict:
        return self.data.get("baud_cfg", {}) or self.data.get("uart_adapt_params", {})

    @property
    def model(self) -> dict:
        return self.extras.get("model", {})

    @property
    def category(self) -> str | None:
        return self.extras.get("category", {})

    @property
    def profile(self) -> dict | None:
        profiles = self.data.get("profiles", [])
        profile = profiles and profiles[0] or {}
        if isinstance(profile, dict):
            return profile
        return None

    @property
    def profile_slug(self) -> str | None:
        profiles = self.data.get("profiles", [])
        profile = profiles and profiles[0] or {}
        if isinstance(profile, str):
            return profile.lower()
        slug = profile.get("slug", None)
        return slug and slug.lower()

    @property
    def chip_name(
        self,
    ) -> Literal["BK7231T"] | Literal["BK7231N"] | Literal["?"] | None:
        module = self.upk.get("module", "")
        if module:
            match module[0:2].upper():
                case "WB":
                    return "BK7231T"
                case "CB":
                    return "BK7231N"
                case _:
                    return "?"
        profile = self.profile
        if profile and "firmware" in profile:
            chip = profile["firmware"].get("chip", None)
            if chip == "BK7231T":
                return "BK7231T"
            if chip == "BK7231N":
                return "BK7231N"
        slug = self.profile_slug
        if slug:
            if "bk7231t" in slug:
                return "BK7231T"
            if "bk7231s" in slug:
                return "BK7231T"
            if "bk7231n" in slug:
                return "BK7231N"
        if self.type == ConfigData.Type.STORAGE:
            if "uart_adapt_params" in self.data:
                return "BK7231T"
            if "baud_cfg" in self.data:
                return "BK7231N"
        return None

    @property
    def schema_id(self) -> str | None:
        match self.type:
            case ConfigData.Type.CLOUDCUTTER:
                return list(self.data.get("schemas", {}) or [None])[0]
            case ConfigData.Type.STORAGE:
                return self.data.get("gw_di", {}).get("s_id", None)

    @property
    def data_device(self) -> dict | None:
        match self.type:
            case ConfigData.Type.CLOUDCUTTER:
                key = self.data.get("key", None)
                return dict(
                    firmwareKey=key if key.startswith("key") else None,
                    productKey=key if not key.startswith("key") else None,
                    factoryPin=None,
                    softwareVer=self.profile.get("firmware", {}).get("version", None),
                )
            case ConfigData.Type.STORAGE:
                gw_di = self.data.get("gw_di", {})
                gw_bi = self.data.get("gw_bi", {})
                pk = gw_di.get("pk", None)
                return dict(
                    firmwareKey=gw_di.get("firmk", None),
                    productKey=pk if pk and not pk.startswith("key") else None,
                    factoryPin=gw_bi.get("fac_pin", None),
                    softwareVer=gw_di.get("swv", None),
                )

    @property
    def data_software(self) -> dict | None:
        if self.type != ConfigData.Type.STORAGE:
            return None
        gw_di = self.data.get("gw_di", {})
        return dict(
            bv=gw_di.get("bv", None),
            pv=gw_di.get("pv", None),
            cadv=gw_di.get("cadv", None),
            cdv=gw_di.get("cdv", None),
        )

    @property
    def data_license(self) -> dict | None:
        if self.type != ConfigData.Type.STORAGE:
            return None
        gw_bi = self.data.get("gw_bi", {})
        if gw_bi.get("ap_ssid", "") == "A":
            return None
        return dict(
            uuid=gw_bi.get("uuid", None),
            authKey=gw_bi.get("auth_key", None),
        )



#  Copyright (c) Kuba Szczodrzyński 2023-9-6.

from dataclasses import dataclass
from enum import Enum
from typing import Literal


@dataclass
class Datapoint:
    id: int
    code: str
    mode: Literal["rw"] | Literal["ro"] | Literal["wo"]
    name: str
    description: str
    spec: "Datapoint.Spec"

    @dataclass
    class Spec:
        type: "Datapoint.Spec.Type"
        max: int = None
        min: int = None
        scale: int = None
        step: int = None
        unit: str = None
        maxlen: int = None
        label: list[str] = None
        range: list[str] = None
        typeDefaultValue: int | bool | str = None

        class Type(Enum):
            BOOL = "bool"
            VALUE = "value"
            ENUM = "enum"
            BITMAP = "bitmap"
            STRING = "string"
            RAW = "raw"

    @staticmethod
    def build(dp: dict) -> "Datapoint":
        spec = dp.get("typeSpec", None)
        spec["type"] = Datapoint.Spec.Type(spec["type"])
        return Datapoint(
            id=dp.get("abilityId"),
            code=dp.get("code"),
            mode=dp.get("accessMode"),
            name=dp.get("name"),
            description=dp.get("description"),
            spec=spec and Datapoint.Spec(**spec),
        )

    @property
    def info(self) -> str:
        return (
            f"DP {self.id}, "
            f"{self.spec.type.name}({self.mode}), "
            f"{self.code} ({self.name})"
        )



#  Copyright (c) Kuba Szczodrzyński 2023-5-13.

from logging import debug, warning
from os import stat
from os.path import isfile
from pprint import pprint
from socket import gethostbyname
from time import sleep
from typing import Callable
from urllib.parse import urlparse

import requests
from bk7231tools.analysis.kvstorage import KVStorage
from ltchiptool.gui.work.base import BaseThread
from ltchiptool.util.flash import ClickProgressCallback


class UpkThread(BaseThread):
    def __init__(
        self,
        file: str = None,
        url: str = None,
        on_storage: Callable[[dict], None] = None,
    ):
        super().__init__()
        self.file = file
        self.url = url
        self.on_storage = on_storage

    def run_impl(self):
        if self.file is not None:
            self.run_file(self.file)
        if self.url is not None:
            self.run_url(self.url)

    # noinspection HttpUrlsUsage
    def run_url(self, url: str):
        # grab storage data from Kickstart API
        if not url.startswith("http"):
            url = "http://" + url
        url = urlparse(url)
        url = url.netloc
        if not url:
            raise ValueError(f"Invalid URL: {url}")

        offset = 0x1E0000
        start = 0x1E0000 - offset
        end = 0x200000 - offset
        init_size = 1024
        block_size = 2048
        buffer = bytearray(end - start)

        with ClickProgressCallback(length=end - start) as bar:
            bar.on_message(f"Resolving {url}...")
            try:
                ip = gethostbyname(url)
            except Exception:
                raise ConnectionError(f"Couldn't find hostname: {url}")

            url = f"http://{ip}/hub/flash_read"
            params = dict(offset=0, length=init_size)

            bar.on_message(f"Connecting to {ip}...")
            with requests.get(url, params, timeout=5.0) as r:
                data = r.content
                if len(data) != init_size:
                    raise RuntimeError(
                        f"Incomplete response read: {len(data)}/{init_size}\n\n"
                        f"Is the chip running Kickstart firmware?"
                    )

            while start < end and self.should_run():
                bar.on_message(f"Reading from 0x{offset + start:06X}")
                read_size = min(block_size, end - start)
                params["offset"] = offset + start
                params["length"] = read_size
                sleep(0.05)
                with requests.get(url, params, timeout=5.0) as r:
                    data = r.content
                    if len(data) != read_size:
                        warning(f"Incomplete response read: {len(data)}/{read_size}")
                        sleep(0.2)
                        continue
                    bar.on_update(read_size)
                    buffer[start : start + read_size] = data
                    start += read_size

        if not self.should_run():
            return

        self.run_data(buffer)

    def run_file(self, file: str):
        # read storage from file
        if not isfile(file):
            raise FileNotFoundError(f"File '{file}' not found")
        size = stat(file).st_size
        if size > 0x400000:
            # file too large
            raise RuntimeError(f"File larger than 4 MiB ({size}), refusing to load!")
        if size == 0x200000:
            # probably full flash dump
            with open(file, "rb") as f:
                # seek to approx. storage start (plus a fix for BkWriter etc.)
                f.seek(0x1E0000 - 0x11000)
                data = f.read()
            self.run_data(data)
            return
        # try to search the entire file
        with open(file, "rb") as f:
            data = f.read()
        self.run_data(data)

    def run_data(self, data: bytes):
        # parse raw storage
        result = KVStorage.find_storage(data)
        if not result:
            raise ValueError("File doesn't contain known storage area")

        _, data = result
        try:
            kvs = KVStorage.decrypt_and_unpack(data)
        except Exception:
            raise RuntimeError("Couldn't unpack storage data - see program logs")

        keys = list(kvs.indexes.keys())
        debug(f"Found {len(keys)} keys! {keys}")
        if not keys:
            # noinspection PyBroadException
            try:
                pprint(kvs)
            except Exception:
                pass
            raise RuntimeError("No keys found in storage! Is the data corrupt?")

        try:
            storage = kvs.read_all_values_parsed()
        except Exception:
            # noinspection PyBroadException
            try:
                pprint(kvs)
            except Exception:
                pass
            raise RuntimeError("Couldn't parse storage data - see program logs")

        self.on_storage(storage)

    def stop(self):
        super().stop()
